//+------------------------------------------------------------------+
//|                                                    SignalQLM.mqh |
//|                   Copyright 2009-2017, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Expert\ExpertSignal.mqh>
#include <My\Cmlp--.mqh>
//+------------------------------------------------------------------+
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Signals based on Q-Learning + Markov for MLP.              |
//| Type=SignalAdvanced                                              |
//| Name=Q-Learning + Markov                                         |
//| ShortName=QLM                                                    |
//| Class=CSignalQLM                                                 |
//| Page=signal_qlm                                                  |
//| Parameter=QL_Scale,int,5, Matrix Row-to-Col Scale                |
//| Parameter=QL_Markov,bool,true, Use Markov                        |
//| Parameter=QL_Epsilon,double,0.1,Epsilon for Controling Exploring vs Exploiting|
//| Parameter=TrainSet,int,65,Training Set Size                      |
//| Parameter=Epochs,int,5,Epochs                                    |
//| Parameter=LearningRate,double,0.0005,Learning Rate               |
//| Parameter=InitialWeight,double,0.005,Default Weight              |
//| Parameter=InitialBias,double,0.0,Default Bias                    |
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| QLMs CSignalQLM.                                                 |
//| Purpose: QLMs of Q-Learning + Markov for MLP.                    |
//|            Derives from class CExpertSignal.                     |
//+------------------------------------------------------------------+
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#define  __MLP_INPUTS 2
#define  __MLP_HIDDEN 8
#define  __MLP_OUTPUTS 3
int      __MLP_ARCH[3] // network arch

=
{  __MLP_INPUTS, __MLP_HIDDEN, __MLP_OUTPUTS
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CSignalQLM   : public CExpertSignal
{
protected:

   int                           m_actions;           // LetMarkov possible actions
   int                           m_environments;      // Environments, per matrix axis
   int                           m_scale;             // Environments, row-to-col scale
   bool                          m_use_markov;        // Use Markov
   double                        m_epsilon;                       // Epsilon


   int                           m_train_set;                       //
   int                           m_epochs;                              // Epochs
   double                        m_learning_rate;                       // Alpha

   Elearning                     m_learning_type;

   double                        m_train_deposit;
   double                        m_train_criteria;

   Eregularize                   m_regularization;
   ENUM_MATRIX_NORM              m_norm;
   double                        m_lambda;

   ENUM_ACTIVATION_FUNCTION      m_activation;
   ENUM_LOSS_FUNCTION            m_loss;

   double                        m_initial_weight;
   double                        m_initial_bias;

   Slearning                     m_learning;
   Smlp                          m_mlp;

public:
   void                          CSignalQLM(void);
   void                          ~CSignalQLM(void);

   //--- methods of setting adjustable parameters
   void                          QL_Scale(int value)
   {  m_scale = value;
   }
   void                          QL_Markov(bool value)
   {  m_use_markov = value;
   }
   void                          QL_Epsilon(double value)
   {  m_epsilon = value;
   }


   void                          TrainSet(int value)
   {  m_train_set = value;
   }
   void                          Epochs(int value)
   {  m_epochs = value;
   }
   void                          LearningRate(double value)
   {  m_learning_rate = value;
   }
   void                          LearningType(Elearning value)
   {  m_learning_type = value;
   }


   void                          ActivationType(ENUM_ACTIVATION_FUNCTION value)
   {  m_activation = value;
   }
   void                          LossType(ENUM_LOSS_FUNCTION value)
   {  m_loss = value;
   }
   void                          InitialWeight(double value)
   {  m_initial_weight = value;
   }
   void                          InitialBias(double value)
   {  m_initial_bias = value;
   }

   void                          RegularizationType(Eregularize value)
   {  m_regularization = value;
   }
   void                          RegularizationNorm(ENUM_MATRIX_NORM value)
   {  m_norm = value;
   }
   void                          RegularizationLambda(double value)
   {  m_lambda = value;
   }

   //--- method of verification of arch
   virtual bool      ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool      InitIndicators(CIndicators *indicators);
   //--- methods of checking if the market models are formed
   virtual int       LongCondition(void);
   virtual int       ShortCondition(void);

protected:
   void                          GetOutput(int &Output);

   Cmlp                          *MLP;
};
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
void CSignalQLM::CSignalQLM(void) :    m_train_set(65),
   m_epochs(5),
   m_learning_rate(0.0005),
   m_initial_weight(0.005),
   m_initial_bias(0.0),
   m_actions(3),
   m_environments(3),
   m_scale(5),
   m_use_markov(true),
   m_epsilon(0.1)

{
//--- initialization of protected data
   m_used_series = USE_SERIES_OPEN + USE_SERIES_HIGH + USE_SERIES_LOW + USE_SERIES_CLOSE + USE_SERIES_SPREAD + USE_SERIES_TIME;
   ArrayResize(m_mlp.arch, ArraySize(__MLP_ARCH));
   for(int i = 0; i < ArraySize(__MLP_ARCH); i++)
   {  m_mlp.arch[i] = __MLP_ARCH[i];
   }
   m_mlp.initial_weight = m_initial_weight;
   m_mlp.initial_bias = m_initial_bias;
   m_mlp.loss_typical = LOSS_CCE;
   m_mlp.activation = AF_SOFTMAX;
   m_mlp.regularization = REGULARIZE_L1;
   m_mlp.regularization_norm = MATRIX_NORM_NUCLEAR;
   m_mlp.regularization_lambda = 1.0;
   m_mlp.loss_custom = LOSS_QL;
   m_mlp.ql.actions = m_actions;
   m_mlp.ql.environments = m_environments;
   m_mlp.ql.use_markov = m_use_markov;
   m_mlp.ql.epsilon = m_epsilon;
//
   MLP = new Cmlp(m_mlp);
//
   m_learning.epochs = m_epochs;
   m_learning.initial_rate = m_learning_rate;
   m_learning.rate = m_learning_rate;
   m_learning.type = LEARNING_FIXED;
   m_train_deposit = AccountInfoDouble(ACCOUNT_BALANCE);
   m_train_criteria = 0.0;
}
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
void CSignalQLM::~CSignalQLM(void)
{  delete MLP;
}
//+------------------------------------------------------------------+
//| Validation arch protected data.                              |
//+------------------------------------------------------------------+
bool CSignalQLM::ValidationSettings(void)
{  if(!CExpertSignal::ValidationSettings())
      return(false);
//--- initial data checks
   if(m_epochs < 1)
   {  printf(__FUNCSIG__ + " epochs size should at least be 1! ");
      return(false);
   }
   if(m_learning_rate < 0.0)
   {  printf(__FUNCSIG__ + " learning rate should be greater than 0.0! ");
      return(false);
   }
   if(!MLP.validated)
   {  printf(__FUNCSIG__ + " invalid network arch. ");
      printf(__FUNCSIG__ +
             " invalid network arch! Settings size is: %i, Max layer size is: %i, Min layer size is: %i, and activation is %s ",
             ArraySize(m_mlp.arch), m_mlp.arch[ArrayMaximum(m_mlp.arch)], m_mlp.arch[ArrayMinimum(m_mlp.arch)], EnumToString(m_mlp.activation)
            );
      return(false);
   }
//read best weights
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CSignalQLM::InitIndicators(CIndicators *indicators)
{
//--- check pointer
   if(indicators == NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertSignal::InitIndicators(indicators))
      return(false);
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| "Voting" that price will grow.                                   |
//+------------------------------------------------------------------+
int CSignalQLM::LongCondition(void)
{  int result = 0;
   int _act;
   GetOutput(_act);
   if(_act == 2)
   {  result = 100;
   }
//printf(__FUNCSIG__ + " output is: %i,  and result is: %i", _act,  result);return(0);
   return(result);
}
//+------------------------------------------------------------------+
//| "Voting" that price will fall.                                   |
//+------------------------------------------------------------------+
int CSignalQLM::ShortCondition(void)
{  int result = 0;
   int _act;
   GetOutput(_act);
   if(_act == 0)
   {  result = 100;printf(__FUNCSIG__);
   }
//printf(__FUNCSIG__ + " output is: %i,  and result is: %i", _act,  result);return(0);
   return(result);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSignalQLM::GetOutput(int &Output)
{  m_learning.rate = m_learning_rate;
   for(int i = m_epochs; i >= 1; i--)
   {  MLP.LearningType(m_learning, i);
      for(int ii = m_train_set; ii >= 0; ii--)
      {  vector _in, _in_row, _in_row_old, _in_col, _in_col_old;
         if
         (
            _in_row.Init(m_scale) &&
            _in_row.CopyRates(m_symbol.Name(), m_period, 8, ii + 1, m_scale) &&
            _in_row.Size() == m_scale
            &&
            _in_row_old.Init(m_scale) &&
            _in_row_old.CopyRates(m_symbol.Name(), m_period, 8, ii + 1 + 1, m_scale) &&
            _in_row_old.Size() == m_scale
            &&
            _in_col.Init(m_scale) &&
            _in_col.CopyRates(m_symbol.Name(), m_period, 8, ii + 1, m_scale) &&
            _in_col.Size() == m_scale
            &&
            _in_col_old.Init(m_scale) &&
            _in_col_old.CopyRates(m_symbol.Name(), m_period, 8, m_scale + ii + 1, m_scale) &&
            _in_col_old.Size() == m_scale
         )
         {  _in_row -= _in_row_old;
            _in_col -= _in_col_old;
            vector _in_e;
            _in_e.Init(m_scale);
            MLP.QL.Environment(_in_row, _in_col, _in_e);
            int _row = 0, _col = 0;
            MLP.QL.SetMarkov(int(_in_e[m_scale - 1]), _row, _col);
            _in.Init(__MLP_INPUTS);
            _in[0] = _row;
            _in[1] = _col;
            MLP.Set(_in);
            MLP.Forward();
            if(ii > 0)
            {  vector _target, _target_data;
               if
               (
                  _target_data.Init(2) &&
                  _target_data.CopyRates(m_symbol.Name(), m_period, 8, ii, 2) &&
                  _target_data.Size() == 2
               )
               {  _target.Init(__MLP_OUTPUTS);
                  _target.Fill(0.0);
                  double _type = _target_data[0] - _in_row[1];
                  int _index = (_type < 0.0 ? 0 : (_type > 0.0 ? 2 : 1));
                  _target[_index] = 1.0;
                  MLP.Get(_target);
                  m_learning.ql_e = _in_e;
                  m_learning.ql_reward_float = _in_row[m_scale - 1];
                  m_learning.ql_reward_max = _in_row.Max();
                  m_learning.ql_reward_min = _in_row.Min();
                  if(i == m_epochs && ii == m_train_set)
                  {  MLP.QL.Action();
                  }
                  MLP.Backward(m_learning, i);
               }
            }
            Output = (MLP.output.Max()==MLP.output[0]?0:(MLP.output.Max()==MLP.output[1]?1:2));
         }
      }
   }
}
//+------------------------------------------------------------------+
